/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

/* 
 * File:   virltc2600.cpp
 * Author: claude
 *
 * Created on 22 décembre 2016, 18:57
 */

#include <cstdlib>

using namespace std;

#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>
#include <locale.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <getopt.h>

#include <curses.h>

WINDOW * wndChip[ 2 ];
WINDOW * wndDAC[ 8 ];
WINDOW * wndLog;
unsigned short dacValues[ 8 ] = {0, 0, 0, 0, 0, 0, 0, 0};
volatile sig_atomic_t receivedCmd = -1;

int nbLTC = -1;

struct {
    int inputRegValue;
    int dacRegValue;
    bool changed;
} channels[ 8 ];

/**
 * 
 * @param progName
 */
void displayUsage(const char * progName) {
    printf("\nusage : %s  -n|--nbltc NBLTC [-h|--help]\n", progName);
    printf("avec :\n");
    printf(". NBLTC : nombre de LTC utilisés pour piloter les 8 voies (1 ou 2)\n");

    exit(1);
}

/**
 * 
 * @param win
 * @param dacNum
 * @param val
 * @param mode
 * @return 
 */
int displayDacWin(int dacIdx, char mode) {
    int y, x;
    int i, j;
    WINDOW * win;
    double height;
    char name[] = "Voie ?";
    char inputRegStr[ 1024 ];
    char dacRegStr[ 1024 ];

    win = wndDAC[ dacIdx ];
    werase(win);
    box(win, 0, 0);
    name[ 5 ] = dacIdx + '1';
    mvwprintw(win, 0, 1, "[ %s ]", name);

    if (mode == 'V') {
        sprintf(inputRegStr, "%08.6f V", (channels[ dacIdx ].inputRegValue >> 4) * 5.0 / 4095);
        sprintf(dacRegStr, "%08.6f V", (channels[ dacIdx ].dacRegValue >> 4) * 5.0 / 4095);
    } else {
        sprintf(inputRegStr, "  0x%04x  ", channels[ dacIdx ].inputRegValue);
        sprintf(dacRegStr, "  0x%04x  ", channels[ dacIdx ].dacRegValue);
    }

    if (channels[ dacIdx ].dacRegValue == -1) {
        sprintf(dacRegStr, mode == 'V' ? "----------" : "  ------  ");
    }

    wattron(win, COLOR_PAIR(1) | A_BOLD);
    mvwaddstr(win, 1, 2, inputRegStr);
    wattroff(win, COLOR_PAIR(1) | A_BOLD);
    wattron(win, COLOR_PAIR(2) | A_BOLD);
    mvwprintw(win, 2, 2, dacRegStr);
    wattroff(win, COLOR_PAIR(2) | A_BOLD);

    wnoutrefresh(win);

    return 0;
}

/**
 * 
 * @param cmd
 * @return 
 */
int updateLogWin(void) {

    struct timeval tv;
    struct tm* ptm;
    char time_string[40];
    long milliseconds;
    static int logLine = 1;

    /* Obtain the time of day, and convert it to a tm struct. */
    gettimeofday(&tv, NULL);
    ptm = localtime(&tv.tv_sec);
    /* Format the date and time, down to a single second. */
    strftime(time_string, sizeof (time_string), "%Y-%m-%d %H:%M:%S", ptm);
    /* Compute milliseconds from microseconds. */
    milliseconds = tv.tv_usec / 1000;
    /* Print the formatted time, in seconds, followed by a decimal point
      and the milliseconds. */
    if (logLine > 8) {
        wscrl(wndLog, 1);
        //        box(wndLog, 0, 0);
        //        mvwaddstr(wndLog, 0, 1, "[ Log ]");
        mvwvline(wndLog, 8, 0, 0, 1);
        mvwvline(wndLog, 8, 55, 0, 1);
        logLine = 8;
    }
    mvwprintw(wndLog, logLine++, 1, "%s.%03ld - CE%d : [%1x|%1x][%02x][%02x]"
            //    wprintw(wndLog, "\n%s.%03ld - [%1x][%1x][%04x] "
            , time_string
            , milliseconds
            , (receivedCmd >> 31) & 0x0001
            , (receivedCmd >> 20) & 0x000f
            , (receivedCmd >> 16) & 0x000f
            , (receivedCmd >> 8) & 0x00ff
            , (receivedCmd) & 0x00ff
            );

    wnoutrefresh(wndLog);
}

/**
 * Signale l'arrivée d'une nouvelle commande.
 * !! ATTENTION !! Le choix a été fait d'ignorer une commande si la précédente
 * n'a pas complètement été traitée par le simulateur => Celui-ci ne
 * reflète donc pas forcément fidèlement la réalité lorsque les commandes sont
 * enchainées trop rapidement (i.e. plus vite qu'il ne faut pour mettre à jour 
 * l'écran).
 * @param unused
 * @param info
 * @param unused_ptr
 */
static void handler(int unused, siginfo_t * info, void * unused_ptr) {

    // SI la dernière commande reçue a été complètement traitée ALORS
    if (receivedCmd == -1) {
        // On signale l'arrivée d'une nouvelle commande
        receivedCmd = info->si_value.sival_int;
    }
}

int processCmd(void) {
    int csSPI = (receivedCmd >> 31) & 0x0001;
    int cmd = (receivedCmd >> 20) & 0x000f;
    int dacId = (receivedCmd >> 16) & 0x000f;
    unsigned short data = receivedCmd & 0x0000ffff;
    int first, last;
    int offset;
    int i;

    if (nbLTC == 1) {
        if (dacId == 15) {
            first = 0;
            last = 7;
        } else {
            first = last = dacId;
        }
        offset = 0;
    } else { // nbLTC = 2
        offset = (csSPI == 0) ? 0 : 4;
        if (dacId == 15) {
            first = offset + 0;
            last = offset + 3;
        } else {
            first = last = offset + dacId;
        }
    }

    switch (cmd) {
        case 0: // Write to Input Register n
            for (i = first; i <= last; i++) {
                channels[ i ].inputRegValue = data;
                channels[ i ].changed = true;
            }
            break;

        case 1: // Update (Power Up) DAC Register n
            for (i = first; i <= last; i++) {
                channels[ i ].dacRegValue = channels[ i ].inputRegValue;
                channels[ i ].changed = true;
            }
            break;

        case 2: // Write to Input Register n, Update (Power Up) All n
            if (dacId != 15) {
                channels[ dacId ].inputRegValue = data;
            }
            for (i = 0; i <= 7; i++) {
                channels[ i ].dacRegValue = channels[ i ].inputRegValue;
                channels[ i ].changed = true;
            }
            break;

        case 3: // Write to and Update (Power Up) n
            for (i = first; i <= last; i++) {
                channels[ i ].inputRegValue = data;
                channels[ i ].dacRegValue = data;
                channels[ i ].changed = true;
            }
            break;

        case 4: // Power Down n
            for (i = first; i <= last; i++) {
                channels[ i ].dacRegValue = -1;
                channels[ i ].changed = true;
            }
            break;

        case 15: // No Operation
            break;

        default:
            break;
    }
}

/**
 * 
 * @return 
 */
int main(int argc, char** argv) {
    int opt;
    char *endptr;
    int screenW;
    int screenH;
    int key = 'V';
    char mode = '?';
    int x, y;

    struct sigaction action;
    action.sa_flags = SA_SIGINFO;
    action.sa_sigaction = handler;
    sigfillset(& action.sa_mask);
    sigaction(SIGRTMIN, & action, NULL);

    setlocale(LC_ALL, "fr_FR.UTF-8");
    
    // Gestion des options de la ligne de commande
    while (1) {
        static struct option long_options[] = {
            {"nbltc", required_argument, 0, 'n'},
            {"help", no_argument, 0, 'h'},
            {0, 0, 0, 0}
        };

        /* getopt_long stores the option index here. */
        int option_index = 0;

        opt = getopt_long(argc, argv, "n:", long_options, &option_index);

        /* Detect the end of the options. */
        if (opt == -1) {
            break;
        }

        switch (opt) {
            case 0:
                if (long_options[option_index].flag != 0)
                    break;
                //printf("option %s", long_options[option_index].name);
                //if (optarg)
                //    printf(" with arg %s", optarg);
                //printf("\n");
                break;

            case 'n':
                //printf("option -n with value `%s'\n", optarg);
                nbLTC = strtol(optarg, NULL, 10);

                if ((nbLTC != 1) && (nbLTC != 2)) {
                    puts("Le nombre de LTC doit être 1 ou 2.");
                    exit(1);
                }
                break;

            case 'h':
            case '?':
                displayUsage(argv[0]);
                break;

            default:
                abort();
        }
    }

    // Gestion de l'IHM

    initscr();
    raw();
    curs_set(FALSE);
    noecho();
    //timeout(10);

    screenW = getmaxx(stdscr);
    screenH = getmaxy(stdscr);

    if ((screenW < 56) || (screenH < 22)) {
        mvaddstr(0, 0, "La taille minimale du terminal doit être de 58 colonnes x 24 lignes pour exécuter le simulateur");
        getch();
        endwin();
        exit(-1);
    }

    if (has_colors()) {
        start_color();
        use_default_colors();
        init_pair(1, COLOR_WHITE, COLOR_BLACK);
        init_pair(2, COLOR_CYAN, COLOR_BLACK);
        init_pair(3, COLOR_MAGENTA, COLOR_BLACK);
    }

    attron(COLOR_PAIR(3) | A_BOLD);
    mvaddstr(0, 7, "=== SIMULATEUR de DAC Linear LTC2600 ===");
    attroff(COLOR_PAIR(3) | A_BOLD);

    refresh();

    if (nbLTC == 1) {
        // Tracer cadre pour LTC unique (newwin(nbLig, nbCols, yOrig, xOrig)
        wndChip[ 0 ] = newwin(2 * 4 + 2, 4 * 14 + 2, 2, 0);
        box(wndChip[ 0 ], 0, 0);
        mvwprintw(wndChip[ 0 ], 0, 1, "[ LTC adressé par CE0 ]");
        wnoutrefresh(wndChip[0]);

        for (int i = 0; i < 8; i++) {
            //wndDAC[ i ] = newwin(4, screenW / 3, (i / 3)*4, (i % 3) * screenW / 3);
            wndDAC[ i ] = derwin(wndChip[0], 4, 14, (i / 4)*4 + 1, (i % 4) * 14 + 1);
        }
    } else { //nbLTC = 2
        wndChip[ 0 ] = newwin(1 * 4 + 2, 4 * 14 + 2, 2, 0);
        box(wndChip[ 0 ], 0, 0);
        mvwprintw(wndChip[ 0 ], 0, 1, "[ LTC adressé par CE0 ]");
        wnoutrefresh(wndChip[0]);

        for (int i = 0; i < 4; i++) {
            //wndDAC[ i ] = newwin(4, screenW / 3, (i / 3)*4, (i % 3) * screenW / 3);
            wndDAC[ i ] = derwin(wndChip[0], 4, 14, 1, i * 14 + 1);
        }

        wndChip[ 1 ] = newwin(1 * 4 + 2, 4 * 14 + 2, (1 * 4 + 2) + 2, 0);
        box(wndChip[ 1 ], 0, 0);
        mvwprintw(wndChip[ 1 ], 0, 1, "[ LTC adressé par CE1 ]");
        wnoutrefresh(wndChip[1]);

        for (int i = 4; i < 8; i++) {
            //wndDAC[ i ] = newwin(4, screenW / 3, (i / 3)*4, (i % 3) * screenW / 3);
            wndDAC[ i ] = derwin(wndChip[1], 4, 14, 1, (i % 4) * 14 + 1);
        }
    }

    wndLog = newwin(10, 58, 14, 0); // newwin(nbLins, nbCols, yOrig, xOrig)
    box(wndLog, 0, 0);
    mvwaddstr(wndLog, 0, 1, "[ Log ]");
    wnoutrefresh(wndLog);
    scrollok(wndLog, TRUE);
    wsetscrreg(wndLog, 1, 8);

    while (key != 'Q') {
        switch (key) {
            case 'H':
            case 'V':
                mode = key;
                for (int i = 0; i < 8; i++) {
                    channels[ i ].changed = true;
                }
                break;

            case ERR: // Réception de signal
                if ((errno == EINTR) && (receivedCmd != -1)) {

                    updateLogWin();

                    processCmd();

                    receivedCmd = -1;
                }
                break;

        }

        for (int i = 0; i < 8; i++) {
            if (channels[ i ].changed) {
                displayDacWin(i, mode);
            }
        }
        doupdate();

        key = toupper(getch());
    }

    echo();
    curs_set(TRUE);
    noraw();
    endwin();
}

